@@ -6,6 +6,7 @@
 #include <cpp11/strings.hpp>
 #include <cpp11/integers.hpp>
 #include <cpp11/list.hpp>
+#include <cpp11/list_of.hpp>
 #include <cpp11/logicals.hpp>
 
 #include <systemfonts.h>
@@ -13,18 +14,18 @@
 
 using namespace cpp11;
 
+namespace textshaping {
 struct Point {
   double x;
   double y;
-
-  Point() : x(0.0), y(0.0) {}
-  Point(double _x, double _y) : x(_x), y(_y) {}
 };
+}
 
 [[cpp11::register]]
 list get_string_shape_c(strings string, integers id, strings path, integers index,
-                        doubles size, doubles res, doubles lineheight, integers align,
-                        doubles hjust, doubles vjust, doubles width, doubles tracking,
+                        list_of<list> features, doubles size, doubles res,
+                        doubles lineheight, integers align, doubles hjust,
+                        doubles vjust, doubles width, doubles tracking,
                         doubles indent, doubles hanging, doubles space_before,
                         doubles space_after);
 
@@ -34,7 +35,7 @@ doubles get_line_width_c(strings string, strings path, integers index, doubles s
 int ts_string_width(const char* string, FontSettings font_info,
                     double size, double res, int include_bearing, double* width);
 int ts_string_shape(const char* string, FontSettings font_info, double size,
-                    double res, std::vector<Point>& loc, std::vector<uint32_t>& id,
+                    double res, std::vector<textshaping::Point>& loc, std::vector<uint32_t>& id,
                     std::vector<int>& cluster, std::vector<unsigned int>& font,
                     std::vector<FontSettings>& fallbacks,
                     std::vector<double>& fallback_scaling);
